"""Provides an interface for sampling knapsack instances.

Examples
--------
Sample a random knapsack instance by sampling from default distributions:
    >>> from pykp.sampler import Sampler
    >>> sampler = Sampler(num_items=5, normalised_capacity=0.6)
    >>> knapsack = sampler.sample()
    >>> len(knapsack.items)
    5

Create a sampler with custom distributions:
    >>> import numpy as np
    >>> sampler = Sampler(
    ...     num_items=5,
    ...     normalised_capacity=0.8,
    ...     weight_dist=(
    ...         np.random.default_rng().normal,
    ...         {"loc": 100, "scale": 10},
    ...     ),
    ...     value_dist=(
    ...         np.random.default_rng().normal,
    ...         {"loc": 50, "scale": 5},
    ...     ),
    ... )
    >>> knapsack = sampler.sample()
    >>> len(knapsack.items)
    5
"""

import numpy as np

from .item import Item
from .knapsack import Knapsack


class Sampler:
    """Generate random knapsack instances.

    Sample knapsack instances by specifying  the number of items, normalised
    capacity, and optionally custom distributions for weights and values.

    Parameters
    ----------
    num_items : int
        Number of items to include in each sampled knapsack instance.
    normalised_capacity : float
        Normalised capacity of the knapsack, defined as the sum of item weights
        divided by the capacity constraint.

    Other Parameters
    ----------------
    weight_dist : str, optional
        Name of the distribution to sample item weights from. Defaults to
        uniform distribution over the half-open interval [0.001, 1).
    weight_dist_kwargs : dict, optional
        Additional keyword arguments to pass to the weight distribution
        function. Defaults to None.
    value_dist : str, optional
        Name of the distribution to sample item values from. Defaults to
        uniform distribution over the half-open interval [0.001, 1).
    value_dist_kwargs : dict, optional
        Additional keyword arguments to pass to the value distribution
        function. Defaults to None.

    .. note::
        Find a list of available distributions in the `numpy.random.Generator`
        documentation:
        https://numpy.org/doc/stable/reference/random/generator.html#distributions.

    Raises
    ------
    ValueError
        If `weight_dist` is specified without providing `weight_dist_kwargs`.
        If `value_dist` is specified without providing `value_dist_kwargs`.

    Examples
    --------
    Sample a random knapsack instance by sampling from default distributions:
        >>> from pykp.sampler import Sampler
        >>> sampler = Sampler(num_items=5, normalised_capacity=0.6)
        >>> knapsack = sampler.sample()
        >>> len(knapsack.items)
        5

    Create a sampler with custom distributions:
        >>> import numpy as np
        >>> sampler = Sampler(
        ...     num_items=5,
        ...     normalised_capacity=0.8,
        ...     weight_dist="normal",
        ...     weight_dist_kwargs={"loc": 100, "scale": 10},
        ...     value_dist="normal",
        ...     value_dist_kwargs={"loc": 50, "scale": 5},
        ... )
        >>> knapsack = sampler.sample()
        >>> len(knapsack.items)
        5
    """

    def __init__(
        self,
        num_items: int,
        normalised_capacity: float,
        weight_dist: str = "uniform",
        value_dist: str = "uniform",
        weight_dist_kwargs: dict | None = None,
        value_dist_kwargs: dict | None = None,
    ):
        self.num_items = num_items
        self.normalised_capacity = normalised_capacity

        if weight_dist != "uniform" and weight_dist_kwargs is None:
            raise ValueError(
                "`weight_dist_kwargs` must be provided "
                "if `weight_dist` is specified."
            )
        if value_dist != "uniform" and value_dist_kwargs is None:
            raise ValueError(
                "`value_dist_kwargs` must be provided "
                "if `value_dist` is specified."
            )
        self.weight_dist = weight_dist
        self.weight_dist_kwargs = weight_dist_kwargs or {
            "low": 0.001,
            "high": 1,
        }
        self.value_dist = value_dist
        self.value_dist_kwargs = value_dist_kwargs or {"low": 0.001, "high": 1}

    def _sample_dist(self, rng, dist_name, dist_kwargs, size):
        """Call desired distribution from the rng.

        Parameters
        ----------
        rng : np.random.Generator
        dist_name : str
            The name of the distribution to sample from.
        dist_kwargs : dict
            Additional kwargs to pass to the distribution.
        size : int
            Number of items to sample.

        Raises
        ------
        ValueError
            If the distribution name is not recognised.

        Returns
        -------
        np.ndarray
            Sampled array of shape (size,).
        """
        if isinstance(dist_name, str):
            # If it’s a recognized distribution name, call via getattr.
            # E.g. rng.uniform(**dist_kwargs, size=size)
            dist_func = getattr(rng, dist_name, None)
            if dist_func is None:
                raise ValueError(f"Unknown distribution name: {dist_name}")
            return dist_func(**{**dist_kwargs, "size": size})
        else:
            raise TypeError(
                "`weight_dist` and `value_dist` must be either a string "
                "or a callable that accepts (rng, size, **kwargs)."
            )

    def sample(self, seed: int = None) -> Knapsack:
        """Generate a random knapsack instance.

        Samples a knapsack instance using the sampling criteria provided to
        the sampler.

        Parameters
        ----------
        seed : int, optional
            Seed for the random sample. Defaults to None.

        Returns
        -------
        Knapsack
            A `Knapsack` object containing the sampled items and capacity
        """
        rng = np.random.default_rng(seed)

        weights = self._sample_dist(
            rng, self.weight_dist, self.weight_dist_kwargs, self.num_items
        )
        profits = self._sample_dist(
            rng, self.value_dist, self.value_dist_kwargs, self.num_items
        )

        items = np.array(
            [Item(profits[i], weights[i]) for i in range(self.num_items)]
        )

        sum_weights = np.sum([item.weight for item in items])
        kp = Knapsack(
            items=items,
            capacity=self.normalised_capacity * sum_weights,
        )
        return kp



"""Metrics for evaluating knapsack instance complexity.

This module provides metrics to evaluate the complexity and phase transition
properties of knapsack problem instances.
"""

import itertools
import time
from typing import Callable, Literal

import numpy as np
import pandas as pd
from tqdm import tqdm

from pykp.item import Item

from . import solvers
from .arrangement import Arrangement

SOLVERS = ["branch_and_bound", "mzn_gecode"]


def _initialise_grid(
    resolution: tuple[int, int],
) -> tuple[np.ndarray, np.ndarray]:
    """
    Create a 2D grid of normalised capacities and profits.

    This function sets up a mesh grid along the x-axis for normalised capacity
    and the y-axis for normalised profit.

    Parameters
    ----------
    resolution : tuple[int, int]
        The desired resolution of the grid in the form (num_capacities,
        num_profits).

    Returns
    -------
    grid : tuple[np.ndarray, np.ndarray]
        The mesh grid arrays representing normalised capacities and profits.
        The first array corresponds to normalised capacities, and the second
        to normalised profits.
    """
    norm_c_step_size = 1 / resolution[1]
    norm_p_step_size = 1 / resolution[0]

    grid = np.meshgrid(
        np.linspace(0, 1 - norm_c_step_size, resolution[1]),
        np.linspace(1 - norm_p_step_size, 0, resolution[0]),
    )

    return grid


def _sample_instance(
    num_items: int, norm_c: float, norm_p: float, rng: np.random.Generator
) -> tuple[list[Item], float, float]:
    """
    Sample a knapsack instance based on normalised capacity and profit targets.

    A set of `num_items` is generated with uniform random weights and values.
    The capacity is defined as the sum of all weights multiplied by `norm_c`,
    and the target profit is defined as the sum of all values multiplied by
    `norm_p`.

    Parameters
    ----------
    num_items : int
        Number of items to generate.
    norm_c : float
        Normalised capacity factor to scale the total weight.
    norm_p : float
        Normalised profit factor to scale the total value.
    rng : np.random.Generator
        Random number generator for sampling.

    Returns
    -------
    items : list[Item]
        The generated knapsack items with random weights and values.
    capacity : float
        The scaled capacity for the knapsack.
    target_profit : float
        The scaled target profit.
    """
    weights = rng.uniform(0.001, 1, num_items)
    profits = rng.uniform(0.001, 1, num_items)

    capacity = sum(weights) * norm_c
    target_profit = sum(profits) * norm_p

    items = [
        Item(value=profits[i], weight=weights[i]) for i in range(num_items)
    ]
    return items, capacity, target_profit


def _simulate_cell_solvability(
    norm_c_range: tuple[float, float],
    norm_p_range: tuple[float, float],
    num_items: int,
    samples: int,
    solver: Callable,
    rng: np.random.Generator,
    progress: tqdm,
) -> float:
    """Estimate the solvability of a grid cell.

    An instance is generated based on a random draw  within `norm_c_range` and
    `norm_p_range`. The instance is solved to check if the optimal solution
    meets or exceeds the target profit threshold. This process is repeated
    `samples` times to estimate the solvability of the cell.

    Parameters
    ----------
    norm_c_range : tuple[float, float]
        Lower and upper bound for the normalised capacity within the grid cell.
    norm_p_range : tuple[float, float]
        Lower and upper bound for the normalised profit within the grid cell.
    num_items : int
        Number of items to generate for each sample.
    samples : int
        Number of sampled instances in the given cell.
    solver : Callable
        A solver function that takes items and capacity as arguments and
        returns an optimal node or array of optimal nodes.
    rng : np.random.Generator
        Random number generator for sampling.
    progress : tqdm
        Progress bar for tracking iterations.

    Returns
    -------
    float
        The fraction of instances in the cell for which the solver found
        a solution that meets or exceeds the target profit.

    """
    total_time = 0
    total_solved = 0
    for _ in range(samples):
        norm_c_draw = rng.uniform(norm_c_range[0], norm_c_range[1])
        norm_p_draw = rng.uniform(norm_p_range[0], norm_p_range[1])

        items, capacity, target_profit = _sample_instance(
            num_items=num_items,
            norm_c=norm_c_draw,
            norm_p=norm_p_draw,
            rng=rng,
        )
        start = time.perf_counter()
        result = solver(items=items, capacity=capacity, target=target_profit)
        end = time.perf_counter()

        total_time += end - start
        total_solved += int(result)
        progress.update(1)

    return (total_solved / samples, total_time / samples)


def _save_phase_transition(
    solvability: np.ndarray,
    time: np.ndarray,
    outcome: str,
    grid: tuple[np.ndarray, np.ndarray],
    path: str,
):
    """Save the phase transition matrix to a CSV file.

    Parameters
    ----------
    phase_transition : np.ndarray
        A 2D array representing the solvability at each grid cell.
    grid : tuple[np.ndarray, np.ndarray]
        The mesh grid of normalised capacities and profits. The first element
        should correspond to normalised capacities, and the second to
        normalised profits.
    outcome : str
        The outcome to save to the CSV file. One of "solvability", "time", or
        "both".
    path : str
        The path (including filename) for the output CSV file.

    """
    data = {
        "nc_lower": grid[0].flatten(),
        "nc_upper": grid[0].flatten() + 1 / len(grid[0][0]),
        "np_lower": grid[1].flatten(),
        "np_upper": grid[1].flatten() + 1 / len(grid[1][0]),
    }

    if outcome == "solvability":
        data["solvability"] = solvability.flatten()
    elif outcome == "time":
        data["time"] = time.flatten()
    else:
        data["solvability"] = solvability.flatten()
        data["time"] = time.flatten()
    df = pd.DataFrame(data)
    df.to_csv(path, index=False, float_format="%.6f")


def phase_transition(
    num_items: int,
    samples: int = 100,
    outcome: str = "both",
    solver: str = "branch_and_bound",
    resolution: tuple[int, int] = (41, 41),
    seed: int | None = None,
    path: str = None,
) -> tuple[np.ndarray, np.ndarray]:
    """Compute the phase transition matrix for knapsack instances.

    In computational problems, a phase transition refers to a phenomenon where
    the probability of finding a solution changes abruptly as some parameter
    crosses a critical threshold. This concept draws inspiration from physics,
    where phase transitions (e.g., water freezing) occur at critical
    thresholds. Such transitions have also been shown to arise in NP-complete
    problems, where the  likelihood of finding a solution changes
    abruptly as a key parameter crosses a critical value.

    For the knapsack problem, this phase transition manifests in terms of
    the relationship between the normalised capacity (`c`) and normalised
    profit (`p`) of an instance. The phase transition has been shown to occur
    when the ratio `r = c / p` is near 1. Instances with 0 < r < 1 are less
    expensive to solve as compred to instances with r ≈ 1.

    This function provide an implementation of the phase transition
    of the knapsack problem, based on Yadav, Nitin, et al. (2018).

    Parameters
    ----------
    num_items : int
        Number of items in each generated knapsack instance.
    samples : int, optional
        Number of instances to evaluate in each cell of the grid. Default is
        100.
    outcome : str, optional
        The outcome to measure in each cell. One or of "solvability" "time"
        or "both". If "solvability" is specified, the phase transition matrix
        represents the proprotion of instances that are satisfiable in each
        cell. If "time" is specified, the phase transition matrix represents
        the average time taken to solve instances in each cell. If "both" is
        specified, the returned phase transition is a tuple with the
        the solvability matrix as the first element, and the time matrix as the
        second element. Default is "both".

    Returns
    -------
        coordinate_matrices : tuple of np.ndarray
            The coordinate matrices for normalised capacities and normalised
            profits. The first matrix corresponds to normalised capacities,
            and the second to normalised profits.
        phase_transition : np.ndarray
            If ``outcome="solvability"``, a 2D matrix where each cell
            represents the proportion of instances that are satisfiable. If
            ``outcome="time"``, a 2D matrix where each cell represents the
            average time taken to solve instances. If ``outcome="both``, a
            tuple containing both matrices.

    Other Parameters
    ----------------
    solver : str, optional
        Which solver to use for evaluating each instance. Default is
        "branch_and_bound".
    resolution: tuple[int, int], optional
        Resolution of the normalised capacity-normalised profit grid.
        The first element corresponds to the resolution of normalised
        profit, and the second to the resolution of normalised capacity.
        Defaults to (41, 41).
    seed : int | None, optional
        Seed for sampling. Defaults to None.
    path str, optional:
        Path to save the phase transition to. Defaults
        to None.

    Examples
    --------
    >>> from pykp.metrics import phase_transition
    >>> grid, (solvability, time) = phase_transition(
    >>>     num_items=12,
    >>>     samples=1000,
    >>>     outcome="both",
    >>>     solver="branch_and_bound",
    >>>     resolution = (20, 20),
    >>> )
    >>> grid[0].shape, grid[1].shape, solvability.shape, time.shape
    ((20, 20), (20, 20), (20, 20), (20, 20))

    Visualise the phase transition matrices using matplotlib:

    >>> import matplotlib.pyplot as plt
    >>> fig, axes = plt.subplots(
    ...     nrows=1,
    ...     ncols=1,
    ...     dpi=200,
    ...     figsize=(4, 3),
    ...     tight_layout=True,
    ... )
    >>> image = axes.imshow(
    ...     phase_transition,
    ...     cmap="RdYlGn",
    ...     interpolation="nearest",
    ...     aspect="auto",
    ...     extent=(0, 1, 0, 1),
    ... )
    >>> axes.set(xlabel="nc", ylabel="np")
    >>> cbar = fig.colorbar(image, ax=axes)
    >>> cbar.ax.set_ylabel("solvability"
    >>> plt.show()

    .. image:: /_static/plots/phase_transition_solvability.png
        :alt: Phase transition solvability matrix

    >>> import matplotlib.pyplot as plt
    >>> fig, axes = plt.subplots(
    ...     nrows=1,
    ...     ncols=1,
    ...     dpi=200,
    ...     figsize=(4, 3),
    ...     tight_layout=True,
    ... )
    >>> image = axes.imshow(
    ...     phase_transition,
    ...     cmap="RdYlGn",
    ...     interpolation="nearest",
    ...     aspect="auto",
    ...     extent=(0, 1, 0, 1),
    ... )
    >>> axes.set(xlabel="nc", ylabel="np")
    >>> cbar = fig.colorbar(image, ax=axes)
    >>> cbar.ax.set_ylabel("solvability"
    >>> plt.show()

    .. image:: /_static/plots/phase_transition_time.png
        :alt: Phase transition time matrix

    >>> # Optionally save results to file:
    >>> phase_transition(
    ...     num_items=10, samples=50, resolution=(5, 5), path="output.csv"
    ... )

    Notes
    -----
    The phase transition is typically applied with reference to the decision
    variant of the knapsack problem. To apply the phase transition to the
    optimisation variant, the target profit should be set to the optimal
    solution value of the instance. One can then compute the normalised
    profit based on the optimal solution, and observe where the instance
    lies in the phase transition matrix.

    References
    ----------

        .. [1] Yadav, Nitin, et al. "Phase transition in the knapsack problem."
           arXiv preprint arXiv:1806.10244 (2018).
    """
    if outcome not in ["solvability", "time", "both"]:
        raise ValueError(
            "`outcome` must one of 'solvability', 'time', or 'both'."
        )

    match solver:
        case "branch_and_bound":
            solver = solvers._branch_and_bound_decision_variant
        case "mzn_gecode":
            solver = solvers._mzn_gecode_decision_variant
        case _:
            raise ValueError(f"`method` must be one of: {SOLVERS}.")

    grid = _initialise_grid(resolution)
    points = list(
        zip(
            [(p, p + 1 / resolution[0]) for p in grid[0].flatten()],
            [(p, p + 1 / resolution[1]) for p in grid[1].flatten()],
        )
    )
    rng = np.random.default_rng(seed)

    phase_transition_solvability = []
    phase_transition_time = []
    with tqdm(total=samples * len(points)) as progress:
        for norm_c_range, norm_p_range in points:
            solvability, time = _simulate_cell_solvability(
                norm_c_range=norm_c_range,
                norm_p_range=norm_p_range,
                num_items=num_items,
                samples=samples,
                solver=solver,
                progress=progress,
                rng=rng,
            )
            phase_transition_solvability.append(solvability)
            phase_transition_time.append(time)

    phase_transition_solvability = np.array(
        phase_transition_solvability
    ).reshape(resolution)
    phase_transition_time = np.array(phase_transition_time).reshape(resolution)

    if path:
        _save_phase_transition(
            solvability=phase_transition_solvability,
            time=phase_transition_time,
            outcome=outcome,
            grid=grid,
            path=path,
        )

    if outcome == "solvability":
        result = phase_transition_solvability
    elif outcome == "time":
        result = phase_transition_time
    else:
        result = (phase_transition_solvability, phase_transition_time)

    return grid, result


def sahni_k(
    arrangement: Arrangement,
    capacity: int,
) -> int:
    """Compute the Sahni-k metric for a given knapsack arrangement.

    The Sahni-k metric [1]_ [2]_ is defined as the smallest number of items
    (k) whose inclusion in the solution is necessary for the greedy algorithm
    (applied to the remaining items) to yield an optimal solution.

    If `k` equals zero, the Sahni-k algorithm coincides with the greedy
    algorithm. If k is equal to the number of items in the solution, the
    algorithm is similar to a brute-force search through the entire search
    space.

    Parameters
    ----------
    arrangement : Arrangement
        A knapsack arrangement (subset of items) for which to compute Sahni-k.
    capacity : int
        The capacity constraint of the knapsack.

    Returns
    -------
        int: Sahni-k value.

    Examples
    --------
    >>> from pykp import Arrangement
    >>> from pykp import Item
    >>> from pykp import metrics
    >>>
    >>> items = [Item(10, 5), Item(20, 8), Item(15, 7)]
    >>> arr = Arrangement(items=items, state=[1, 0, 1])
    >>> metrics.sahni_k(arr, capacity=15)
    2

    References
    ----------

        .. [1] Sahni, Sartaj. "Approximate algorithms for the 0/1 knapsack
           problem." Journal of the ACM (JACM) 22.1 (1975): 115-124.

        .. [2] Murawski, Carsten, and Peter Bossaerts. "How humans solve
           complex problems: The case of the knapsack problem." Scientific
           reports 6.1 (2016): 34851.

    """
    if not isinstance(arrangement, Arrangement):
        raise ValueError("`arrangement` must be of type `Arrangement`.")
    if arrangement.weight > capacity:
        raise ValueError(
            """The total weight of items included in the `Arrangement` exceeds 
            the `capacity`."""
        )

    in_items = [
        arrangement.items[i]
        for i, element in enumerate(arrangement.state)
        if element == 1
    ]
    for subset_size in range(0, len(arrangement.state) + 1):
        for subset in itertools.combinations(in_items, subset_size):
            subset = list(subset)
            weight = sum([item.weight for item in subset])

            # Solve greedily
            while True:
                if len(subset) == len(arrangement.items):
                    break

                # Check instance at capacity
                out_items = [
                    item for item in arrangement.items if item not in subset
                ]
                if (
                    min([weight + item.weight for item in out_items])
                    > capacity
                ):
                    break

                densities = [item.value / item.weight for item in out_items]
                max_density_item = out_items[densities.index(max(densities))]
                subset.append(max_density_item)
                weight = sum([item.weight for item in subset])

            if set(subset) == set(in_items):
                return subset_size



"""PyKP: A Python library for the Knapsack Problem."""

from .arrangement import Arrangement
from .item import Item
from .knapsack import Knapsack
from .metrics import phase_transition, sahni_k
from .sampler import Sampler
from .solvers import branch_and_bound, greedy, mzn_gecode

__all__ = [
    "Arrangement",
    "Item",
    "Knapsack",
    "Sampler",
    "phase_transition",
    "sahni_k",
    "branch_and_bound",
    "greedy",
    "mzn_gecode",
]



"""Tests for pykp.sampler module."""

from unittest.mock import MagicMock

import numpy as np
import pytest

from pykp.sampler import Sampler


def test_sampler_init_with_defaults():
    """Test Sampler initialisation with default distributions."""
    sampler = Sampler(num_items=5, normalised_capacity=0.5)
    assert sampler.num_items == 5
    assert sampler.normalised_capacity == 0.5
    assert sampler.weight_dist == "uniform"
    assert isinstance(sampler.weight_dist_kwargs, dict)
    assert "low" in sampler.weight_dist_kwargs
    assert "high" in sampler.weight_dist_kwargs

    assert sampler.value_dist == "uniform"
    assert isinstance(sampler.value_dist_kwargs, dict)
    assert "low" in sampler.value_dist_kwargs
    assert "high" in sampler.value_dist_kwargs


def test_sampler_init_with_custom_distributions():
    """Test Sampler initialisation with custom distributions."""
    custom_weight_dist = (
        "normal",
        {"loc": 10, "scale": 5},
    )
    custom_value_dist = (
        "normal",
        {"loc": 20, "scale": 2},
    )
    num_items = 10
    nc = 0.5

    sampler = Sampler(
        num_items=num_items,
        normalised_capacity=nc,
        weight_dist=custom_weight_dist[0],
        weight_dist_kwargs=custom_weight_dist[1],
        value_dist=custom_value_dist[0],
        value_dist_kwargs=custom_value_dist[1],
    )

    assert sampler.num_items == num_items
    assert sampler.normalised_capacity == nc

    assert sampler.weight_dist == custom_weight_dist[0]
    assert sampler.weight_dist_kwargs == custom_weight_dist[1]
    assert sampler.value_dist == custom_value_dist[0]
    assert sampler.value_dist_kwargs == custom_value_dist[1]


def test_sampler_sample_returns_knapsack():
    """Test that the `sample` method returns a Knapsack instance."""
    sampler = Sampler(num_items=5, normalised_capacity=0.8)
    knapsack = sampler.sample()

    # Check that `knapsack` has the attributes we expect
    # (depending on how your Knapsack class is implemented, these may differ).
    assert hasattr(knapsack, "items")
    assert hasattr(knapsack, "capacity")
    assert len(knapsack.items) == 5


def test_sampler_sample_capacity_calculation():
    """Test that the capacity of the knapsack is calculated correctly."""
    sampler = Sampler(
        num_items=10,
        normalised_capacity=0.5,
        weight_dist="integers",
        weight_dist_kwargs={"low": 1, "high": 2},
        value_dist="integers",
        value_dist_kwargs={"low": 1, "high": 2},
    )
    knapsack = sampler.sample()
    knapsack.items

    assert knapsack.capacity == int(0.5 * 10)  # Should be 4


@pytest.mark.parametrize("normalised_capacity", [0.0, 0.1, 0.9, 1.5])
def test_sampler_different_normalised_capacities(normalised_capacity):
    """Test that the capacity of the knapsack is calculated correctly."""
    sampler = Sampler(num_items=3, normalised_capacity=normalised_capacity)
    knapsack = sampler.sample()
    total_weight = sum(item.weight for item in knapsack.items)
    expected_capacity = normalised_capacity * total_weight

    assert np.isclose(knapsack.capacity, expected_capacity)
    assert len(knapsack.items) == 3


@pytest.mark.parametrize("seed", [1, 2, 3, 4, 5])
def test_sample_reporducible(seed: int):
    """Test that the samples are reproducible."""
    sampler = Sampler(num_items=5, normalised_capacity=0.8)

    knapsack = sampler.sample(seed=seed)
    knapsack2 = sampler.sample(seed=seed)
    values = [item.value for item in knapsack.items]
    values2 = [item.value for item in knapsack2.items]
    weights = [item.weight for item in knapsack.items]
    weights2 = [item.weight for item in knapsack2.items]

    assert knapsack.capacity == knapsack2.capacity
    assert np.array_equal(values, values2)
    assert np.array_equal(weights, weights2)
    assert np.array_equal(knapsack.state, knapsack2.state)



"""Implementations of various solvers for the knapsack problem."""

"""
Provides an implementation of branch and bound algorithm for
solving the knapsack problem.

Example:
    To solve a knapsack problem instance using the branch-and-bound algorithm,
    first create a list of items and then call the solver with the items and
    capacity::

        from pykp import Item, Solvers

        items = [
            Item(value=10, weight=5),
            Item(value=15, weight=10),
            Item(value=7, weight=3),
        ]
        capacity = 15
        optimal_nodes = solvers.branch_and_bound(items, capacity)
        print(optimal_nodes)

    Alternatively, construct an instance of the `Knapsack` class and call the
    `solve` method with "branch_and_bound" as the `method` argument::

        from pykp import Item, Knapsack

        items = [
            Item(value=10, weight=5),
            Item(value=15, weight=10),
            Item(value=7, weight=3),
        ]
        capacity = 15
        instance = Knapsack(items=items, capacity=capacity)
        optimal_nodes = instance.solve(method="branch_and_bound")
        print(optimal_nodes)
"""

import itertools
import operator
from collections import defaultdict
from dataclasses import dataclass, field
from queue import PriorityQueue
from typing import Tuple

import nest_asyncio
import numpy as np
from minizinc import Instance, Model, Solver

from .arrangement import Arrangement
from .item import Item


@dataclass(order=True, frozen=True)
class Node:
    """Represents a node in the branch-and-bound tree.

    Parameters
    ----------
    priority : float
        The priority of the node.
    upper_bound : float
        The upper bound of the node.
    items : np.ndarray[Item]
        Items that can be included in the knapsack.
    value : int
        The total value of items in the node.
    weight : int
        The total weight of items in the node.
    included_items : np.ndarray[Item]
        Items included by this node.
    excluded_items : np.ndarray[Item]
        Items excluded by this node.
    """

    priority: float = field(compare=True)
    upper_bound: float = field(compare=False)
    items: np.ndarray[Item] = field(compare=False)
    value: int = field(compare=False)
    weight: int = field(compare=False)
    included_items: np.ndarray[Item] = field(compare=False)
    excluded_items: np.ndarray[Item] = field(compare=False)


def _calculate_upper_bound(
    items: np.ndarray[Item],
    capacity: int,
    included_items: np.ndarray[Item],
    excluded_items: np.ndarray[Item],
) -> float:
    """Calculate the upper bound of the supplied branch.

    The upper bound is calculated by filling the fractional knapsack with
    items in descending order of value-to-weight ratio.

    Parameters
    ----------
    items: np.ndarray[Item]
        Items that can be included in the knapsack.
    capacity: int
        Maximum weight capacity of the knapsack.
    included_items: np.ndarray[Item]
        Items included by all nodes within the branch.
    excluded_items: np.ndarray[Item]
        Items excluded by all nodes within the branch.

    Returns
    -------
    float
        Upper bound of the branch.
    """
    arrangement = Arrangement(
        items=items,
        state=np.array([int(item in included_items) for item in items]),
    )
    candidate_items = np.array(
        sorted(
            set(items) - set(included_items) - set(excluded_items),
            key=lambda item: item.value / item.weight,
            reverse=True,
        )
    )
    balance = capacity - arrangement.weight

    if balance < 0:
        return -1

    if len(candidate_items) == 0 or balance == 0:
        return arrangement.value

    i = 0
    upper_bound = arrangement.value
    while balance > 0 and i < len(candidate_items):
        item = candidate_items[i]
        added_weight = min(balance, item.weight)
        upper_bound = upper_bound + added_weight * item.value / item.weight
        balance = balance - added_weight
        i += 1
    return upper_bound


def _expand_node(
    node: Node,
    capacity: int,
    incumbent: float,
) -> np.ndarray:
    """Expand a node in the branch-and-bound tree.

    The node is expanded by generating two children nodes: one that includes
    the next item in the knapsack and one that excludes it. The children are
    only returned if the upper bound of the child is greater than or equal to
    the incumbent value.

    Parameters
    ----------
    node: Node
        Node to expand.
    capacity: int
        Maximum weight capacity of the knapsack.
    incumbent: float
        The best value found so far.

    Returns
    -------
    np.ndarray
        The child nodes of the expanded node.
    """
    arrangement = Arrangement(
        items=node.items,
        state=np.array(
            [int(item in node.included_items) for item in node.items]
        ),
    )
    if arrangement.weight > capacity:
        return []

    if len(node.included_items) + len(node.excluded_items) >= len(node.items):
        return []  # No further branching possible

    next_item = node.items[len(node.included_items) + len(node.excluded_items)]

    # Generate children (left-branch includes item, right-branch excludes item)
    # only return them if we do not prune by upper_bound.
    children = []

    for included in [True, False]:
        included_items = (
            np.append(node.included_items, next_item)
            if included
            else node.included_items
        )
        excluded_items = (
            np.append(node.excluded_items, next_item)
            if not included
            else node.excluded_items
        )
        upper_bound = _calculate_upper_bound(
            items=node.items,
            capacity=capacity,
            included_items=included_items,
            excluded_items=excluded_items,
        )
        child = Node(
            priority=-upper_bound,
            items=node.items,
            value=node.value + next_item.value * included,
            weight=node.weight + next_item.weight * included,
            included_items=included_items,
            excluded_items=excluded_items,
            upper_bound=upper_bound,
        )
        if child.upper_bound >= incumbent:
            children.append(child)

    return children


def _is_leaf_node(node: Node, capacity: int) -> bool:
    """Whether a provided node is a leaf node.

    A node is considered a leaf node if the balance is under capacity,
    and all items in the branch have been either included or excluded.

    Parameters
    ----------
    node: Node
        Node to check.
    capacity: int
        Maximum weight capacity of the knapsack.

    Returns
    -------
    bool
        Wheter the node is a leaf node.
    """
    weight = sum([i.weight for i in node.included_items])
    balance = capacity - weight
    if balance < 0:
        return False
    remaining_items = (
        set(node.items) - set(node.included_items) - set(node.excluded_items)
    )
    return len(remaining_items) == 0


def branch_and_bound(
    items: list[Item],
    capacity: float,
    n=1,
) -> list[Arrangement]:
    """Solves the knapsack problem using the branch-and-bound algorithm.

    Parameters
    ----------
    items: list[Item]
        Items that can be included in the knapsack.
    capacity: float
        Maximum weight capacity of the knapsack.

    Other Parameters
    ----------------
    n: int, optional
        The n-best solutions to return. If set to 1, the solver returns all
        solutions that achieve the distinct optimal value. If set to n, the
        solver returns the solutions that achieve the n-highest possible
        values. Defaults to 1.

    Returns
    -------
        list[Arrangement]: The optimal arrangements of items in the
            knapsack.

    Examples
    --------
    Solve a knapsack problem using the branch-and-bound algorithm

    >>> from pykp import Item
    >>> from pykp import solvers
    >>>
    >>> items = [
    ...     Item(value=10, weight=5),
    ...     Item(value=15, weight=10),
    ...     Item(value=5, weight=5),
    >>> ]
    >>> capacity = 15
    >>> solvers.branch_and_bound(items, capacity)
    [(v: 25, w: 15, s: 6)]

    Alternatively, construct an instance of the ``Knapsack`` class and call the
    ``solve`` method with "branch_and_bound" as the ``method`` argument

    >>> from pykp import Item
    >>> from pykp import Knapsack
    >>>
    >>> items = [
    ...     Item(value=10, weight=5),
    ...     Item(value=15, weight=10),
    ...     Item(value=5, weight=5),
    ... ]
    >>> capacity = 15
    >>> instance = Knapsack(items=items, capacity=capacity)
    >>> instance.solve(method="branch_and_bound")
    >>> instance.optimal_nodes
    [(v: 25, w: 15, s: 6)]

    If there are multiple solutions with the same optimal value, all will be
    returned.

    >>> from pykp import Item
    >>> from pykp import Knapsack
    >>>
    >>> items = [
    ...     Item(value=10, weight=5),
    ...     Item(value=15, weight=10),
    ...     Item(value=4, weight=2),
    ...     Item(value=6, weight=3),
    ... ]
    >>> capacity = 15
    >>> instance = Knapsack(items=items, capacity=capacity)
    >>> instance.solve(method="branch_and_bound")
    >>> instance.optimal_nodes
    [(v: 25, w: 15, s: 9), (v: 25, w: 15, s: 7)]

    Use the optional ``n`` argument to return the n-best solutions found by
    the solver.

    >>> from pykp import Item
    >>> from pykp import Knapsack
    >>> from pykp import solvers
    >>>
    >>> items = [
    ...     Item(value=10, weight=5),
    ...     Item(value=15, weight=10),
    ...     Item(value=4, weight=2),
    ...     Item(value=6, weight=3),
    ... ]
    >>> capacity = 15
    >>> solvers.branch_and_bound(items, capacity, n=3)
    [(v: 25, w: 15, s: 9), (v: 25, w: 15, s: 7), (v: 21, w: 13, s: 3)]

    .. note::
        The ``n`` argument is on solution values, not the number of
        solutions. If ``n`` is set to 1, the solver returns all solutions
        that achieve the distinct optimal value. More than one solution
        may be returned if there are multiple solutions with the same
        optimal value. Similarly, if ``n`` is set to `n`, the solver returns
        all solutions that achieve the `n`-highest possible values.
    """
    if len(items) == 0:
        return np.array([Arrangement(items=items, state=np.array([]))])

    items = np.array(
        sorted(items, key=lambda item: item.value / item.weight, reverse=True)
    )
    upper_bound = _calculate_upper_bound(
        items=items,
        capacity=capacity,
        included_items=np.array([]),
        excluded_items=np.array([]),
    )
    root = Node(
        priority=-sum([item.value for item in items]),
        items=items,
        value=0,
        weight=0,
        included_items=np.array([]),
        excluded_items=np.array([]),
        upper_bound=upper_bound,
    )
    queue = PriorityQueue()
    queue.put(root)
    incumbent = 0
    nodes = []
    n_best_values = [0]

    while not queue.empty():
        next = queue.get()
        children = _expand_node(next, capacity, incumbent)
        for child in children:
            if child.upper_bound < incumbent:
                continue

            queue.put(child)
            if child.value >= incumbent and _is_leaf_node(child, capacity):
                n_best_values.append(child.value)
                n_best_values = sorted(n_best_values, reverse=True)[:n]
                incumbent = n_best_values[-1]
                nodes.append(child)

    nodes = [node for node in nodes if node.value >= incumbent]
    result = [
        Arrangement(
            items=items,
            state=np.array(
                [int(item in node.included_items) for item in items]
            ),
        )
        for node in nodes
    ]

    return result


def _branch_and_bound_decision_variant(
    items: list[Item], capacity: float, target: float
) -> bool:
    """Solves the knapsack decision variant using branch-and-bound.

    Parameters
    ----------
    items : list[Item]
        Items that can be included in the knapsack.
    capacity : int
        Maximum weight capacity of the knapsack.
    target : float
        The target value to achieve.

    Returns
    -------
    bool
        Whether the target value can be achieved.
    """
    if len(items) == 0:
        return False

    items = np.array(
        sorted(items, key=lambda item: item.value / item.weight, reverse=True)
    )
    upper_bound = _calculate_upper_bound(
        items=items,
        capacity=capacity,
        included_items=np.array([]),
        excluded_items=np.array([]),
    )
    root = Node(
        priority=-sum([item.value for item in items]),
        items=items,
        value=0,
        weight=0,
        included_items=np.array([]),
        excluded_items=np.array([]),
        upper_bound=upper_bound,
    )
    queue = PriorityQueue()
    queue.put(root)

    while not queue.empty():
        next = queue.get()
        children = _expand_node(next, capacity, target)
        for child in children:
            queue.put(child)
            if child.value >= target:
                return True

    return False


def mzn_gecode(items: list[Item], capacity: float) -> Arrangement:
    """Solves the knapsack problem using the MiniZinc Gecode solver.

    Parameters
    ----------
    items : list[Item]
        Array of items to consider for the knapsack.
    capacity : float
        Maximum weight capacity of the knapsack.

    Returns
    -------
    Arrangement
        The optimal arrangement of items in the knapsack.

    Examples
    --------
    Solve a knapsack problem instance using MiniZinc and Gecode:

    >>> from pykp import Item, solvers
    >>> items = np.array(
    ...     [
    ...         Item(value=10, weight=5),
    ...         Item(value=15, weight=10),
    ...         Item(value=5, weight=5),
    ...     ]
    ... )
    >>> capacity = 15
    >>> solvers.mzn_gecode(items, capacity)
    [(v: 25, w: 15, s: 6)]

    Alternatively, construct an instance of the ``Knapsack`` class and call the
    ``solve`` method with "mzn_geocode" as the ``method`` argument

    >>> from pykp import Item
    >>> from pykp import Knapsack
    >>>
    >>> items = [
    ...     Item(value=10, weight=5),
    ...     Item(value=15, weight=10),
    ...     Item(value=5, weight=5),
    ... ]
    >>> capacity = 15
    >>> instance = Knapsack(items=items, capacity=capacity)
    >>> instance.solve(method="mzn_geocode")
    >>> instance.optimal_nodes
    [(v: 25, w: 15, s: 6)]

    .. note::
        You should have MiniZinc 2.5.0 (or higher) installed on your system to
        use this solver. Refer to the `MiniZinc documentation
        <https://docs.minizinc.dev/en/stable/installation.html>`_
        for installation instructions.

    .. note::
        The MiniZinc Gecode solver is not robust to multiple solutions, and
        will report only the first optimal solution found. If knowing all
        optimal solutions is important, consider using the branch-and-bound
        solver.
    """
    nest_asyncio.apply()
    model = Model()
    model.add_string(
        """
		int: n; % number of objects
		set of int: OBJ = 1..n;
		float: capacity;
		array[OBJ] of float: profit;
		array[OBJ] of float: size;

		%var set of OBJ: x;
		array[OBJ] of var 0..1: x;
		var float: P=sum(i in OBJ)(profit[i]*x[i]);

		constraint sum(i in OBJ)(size[i]*x[i]) <= capacity;

		solve :: int_search(x, first_fail, indomain_max, complete) maximize P;
		"""
    )
    gecode = Solver.lookup("gecode")

    instance = Instance(gecode, model)
    instance["n"] = len(items)
    instance["capacity"] = capacity
    instance["profit"] = [item.value for item in items]
    instance["size"] = [item.weight for item in items]

    result = instance.solve()

    return Arrangement(items=items, state=np.array(result["x"]))


def _mzn_gecode_decision_variant(
    items: list[Item], capacity: float, target: float
) -> Arrangement:
    """Solves the knapsack decision variant using MiniZinc and Gecode.

    Parameters
    ----------
    items : list[Item]
        Array of items to consider for the knapsack.
    capacity : float
        Maximum weight capacity of the knapsack.
    target : float
        The target value to achieve.

    Returns
    -------
    Arrangement
        The optimal arrangement of items in the knapsack.
    """
    nest_asyncio.apply()
    model = Model()
    model.add_string(
        """
        int: n;
        float: capacity;
        float: target;
        array[1..n] of float: size;
        array[1..n] of float: profit;

        array[1..n] of var 0..1: x;

        constraint sum(i in 1..n)(size[i]*x[i]) <= capacity;
        constraint sum(i in 1..n)(profit[i]*x[i]) >= target;

        solve satisfy;
        """
    )
    gecode = Solver.lookup("gecode")

    instance = Instance(gecode, model)
    instance["n"] = len(items)
    instance["capacity"] = capacity
    instance["profit"] = [item.value for item in items]
    instance["size"] = [item.weight for item in items]
    instance["target"] = target

    result = instance.solve()

    return result.status.has_solution()


def greedy(items: list[Item], capacity: int) -> Arrangement:
    """Appy the greedy algorithm to a knapsack problem instance.

    Parameters
    ----------
    items : np.ndarray[Item]
        Array of items to consider for the knapsack.
    capacity : int
        Maximum weight capacity of the knapsack.

    Returns
    -------
    Arrangement
        The greedy arrangement of items in the knapsack.

    Examples
    --------
    Solve a knapsack problem using the greedy algorithm:

    >>> from pykp import Item
    >>> from pykp import solvers
    >>> items = np.array(
    ...     [
    ...         Item(value=100, weight=50),
    ...         Item(value=200, weight=100),
    ...         Item(value=400, weight=300),
    ...     ]
    ... )
    >>> capacity = 300
    >>> solvers.greedy(items, capacity)
    (v: 300, w: 150, s: 6)

    .. note::
        The greedy algorithm is not guaranteed to find the optimal solution
        to the knapsack problem. It is a heuristic algorithm that selects
        the best item at each step based on the value-to-weight ratio,
        until no more items can be added to the knapsack.
    """
    items = np.array(items)
    state = np.zeros(len(items))
    weight = 0
    balance = capacity
    while balance > 0:
        remaining_items = [
            items[i]
            for i, element in enumerate(state)
            if element == 0 and items[i].weight + weight <= capacity
        ]
        if len(remaining_items) == 0:
            break
        best_item = max(
            remaining_items, key=lambda item: item.value / item.weight
        )
        state[items.tolist().index(best_item)] = 1
        balance -= best_item.weight
        weight += best_item.weight

    return Arrangement(items=items, state=state)


def _is_subset_feasible(subset: list[Item], capacity) -> bool:
    """Determine whether subset of items is feasible.

    A subset of items is considered feasible if the total weight of the
    items is less than or equal to the capacity of the knapsack.

    Parameters
    ----------
    subset : list[Item]
        Subset of items.
    capacity : int
        Capacity of the knapsack.

    Returns
    -------
    bool
        Whether the node is feasible.
    """
    weight = sum([i.weight for i in subset])
    balance = capacity - weight
    if balance < 0:
        return False
    return True


def _is_subset_terminal(
    subset: list[Item], items: list[Item], capacity
) -> bool:
    """Determine whether subset of items is terminal.

    A subset of items is considered terminal if the total weight of the
    items is less than or equal to the capacity of the knapsack and no
    remaining items can be added to the knapsack without exceeding the
    capacity.

    Parameters
    ----------
    subset : list[Item]
        Subset of items.
    items : list[Item]
        All items in the knapsack.
    capacity : int
        Capacity of the knapsack.

    Returns
    -------
    bool
        Whether the node is terminal
    """
    weight = sum([i.weight for i in subset])
    balance = capacity - weight
    if balance < 0:
        return False
    remaining_items = set(items) - set(subset)
    for i in remaining_items:
        if i.weight <= balance:
            return False
    return True


def brute_force(
    items: list[Item], capacity: int
) -> Tuple[list, list, list, list]:
    """Solves the knapsack problem using brute force.

    Parameters
    ----------
    items : list[Item]
        List of items to consider for the knapsack.
    capacity : int
        Maximum weight capacity of the knapsack.

    Returns
    -------
    list
        Optimal nodes in the knapsack problem.
    list
        Terminal nodes in the knapsack problem.
    list
        Feasible nodes in the knapsack problem.
    list
        All nodes in the knapsack problem.

    Examples
    --------
    To solve a knapsack problem instance using the brute-force
    algorithm, first create a list of items and then call the solver
    with the items and capacity.

    >>> from pykp import Item, solvers
    >>>
    >>> items = [
    ...     Item(value=10, weight=5),
    ...     Item(value=15, weight=10),
    ...     Item(value=5, weight=5),
    ... ]
    >>> capacity = 15
    >>> (optimal, terminal, feasible, all) = solvers.brute_force(
    ...     items, capacity
    ... )
    >>> print(optimal)
    [(v: 25, w: 15, s: 6)]

    Alternatively, construct an instance of the `Knapsack` class and
    call the `solve` method with "brute_force" as the `method`
    argument. When called in this way, the solver will populate the
    `optimal_nodes`, `terminal_nodes`, `feasible_nodes`, and `nodes`
    attributes of the `Knapsack` instance.

    >>> from pykp import Item, Knapsack
    >>>
    >>> items = [
    ...     Item(value=10, weight=5),
    ...     Item(value=15, weight=10),
    ...     Item(value=5, weight=5),
    ... ]
    >>> capacity = 15
    >>> instance = Knapsack(items=items, capacity=capacity)
    >>>
    >>> instance.solve(method="brute_force")
    >>> instace.optimal_nodes
    [(v: 25, w: 15, s: 6)]
    >>> instance.terminal_nodes
    [(v: 25, w: 15, s: 6), (v: 20, w: 15, s: 3), (v: 15, w: 10, s: 5)]
    >>> instance.feasible_nodes
    [(v: 0, w: 0, s: 0),
    (v: 5, w: 5, s: 1),
    (v: 10, w: 5, s: 4),
    (v: 15, w: 10, s: 2),
    (v: 15, w: 10, s: 5),
    (v: 20, w: 15, s: 3),
    (v: 25, w: 15, s: 6)]
    >>> instance.nodes
    [(v: 10, w: 5, s: 4),
    (v: 15, w: 10, s: 2),
    (v: 5, w: 5, s: 1),
    (v: 25, w: 15, s: 6),
    (v: 15, w: 10, s: 5),
    (v: 20, w: 15, s: 3),
    (v: 30, w: 20, s: 7),
    (v: 0, w: 0, s: 0)]

    .. note::
        The brute-force algorithm is computationally expensive and should be
        used with caution for large problem instances.
    """
    nodes = np.array([])
    feasible_nodes = np.array([])
    terminal_nodes = np.array([])
    optimal_nodes = np.array([])

    for i in range(1, len(items) + 1):
        subsets = list(itertools.combinations(items, i))
        for subset in subsets:
            nodes = np.append(
                nodes,
                Arrangement(
                    items=items,
                    state=np.array([int(item in subset) for item in items]),
                ),
            )
            if _is_subset_feasible(subset, capacity):
                feasible_nodes = np.append(
                    feasible_nodes,
                    Arrangement(
                        items=items,
                        state=np.array(
                            [int(item in subset) for item in items]
                        ),
                    ),
                )
            if _is_subset_terminal(subset, items, capacity):
                terminal_nodes = np.append(
                    terminal_nodes,
                    Arrangement(
                        items=items,
                        state=np.array(
                            [int(item in subset) for item in items]
                        ),
                    ),
                )
    nodes = np.append(
        nodes,
        Arrangement(items=items, state=np.zeros(len(items), dtype=int)),
    )
    feasible_nodes = np.append(
        feasible_nodes,
        Arrangement(items=items, state=np.zeros(len(items), dtype=int)),
    )
    feasible_nodes = sorted(
        feasible_nodes,
        key=operator.attrgetter("value"),
    )
    terminal_nodes = sorted(
        terminal_nodes, key=operator.attrgetter("value"), reverse=True
    )
    optimal_nodes = np.array(
        [
            arrangement
            for arrangement in terminal_nodes
            if arrangement.value == terminal_nodes[0].value
        ]
    )
    return (
        list(optimal_nodes),
        list(terminal_nodes),
        list(feasible_nodes),
        list(nodes),
    )



"""Tests for pykp.solvers module."""

import json

import numpy as np
import pytest

from pykp import Item, Sampler, solvers

HEURISTIC_SOLVERS = ["greedy"]
OPTIMAL_SOLVERS = [
    "branch_and_bound",
    # "mzn_geocode",
]
ALL_SOLVERS = HEURISTIC_SOLVERS + OPTIMAL_SOLVERS

with open("tests/test_cases.json") as f:
    TEST_CASES = json.load(f)


@pytest.fixture
def solver(request):
    """Define the solver to be used in the test."""
    if request.param == "greedy":
        solver = solvers.greedy
    elif request.param == "branch_and_bound":
        solver = solvers.branch_and_bound
    elif request.param == "mzn_geocode":
        solver = solvers.mzn_gecode

    return solver


@pytest.mark.parametrize("solver", ALL_SOLVERS, indirect=True)
def test_empty_items(solver):
    """Test the case where there are no items."""
    items = np.array([])
    solutions = solver(items, 0)

    if not isinstance(solutions, np.ndarray):
        solutions = [solutions]

    assert len(solutions) == 1
    assert solutions[0].value == 0
    assert solutions[0].weight == 0


@pytest.mark.parametrize("solver", ALL_SOLVERS, indirect=True)
def test_single_item_fits(solver):
    """Test a single item that fits in the knapsack."""
    items = np.array([Item(value=10, weight=5)])
    capacity = 10
    solutions = solver(items, capacity)

    if not isinstance(solutions, list):
        solutions = [solutions]

    assert len(solutions) == 1
    assert solutions[0].value == 10
    assert solutions[0].weight == 5


@pytest.mark.parametrize("solver", ALL_SOLVERS, indirect=True)
def test_single_item_does_not_fit(solver):
    """Test a single item that does not fit in the knapsack."""
    items = np.array([Item(value=10, weight=15)])
    capacity = 10
    solutions = solver(items, capacity)

    if not isinstance(solutions, list):
        solutions = [solutions]

    assert len(solutions) == 1
    assert solutions[0].value == 0
    assert solutions[0].weight == 0


@pytest.mark.parametrize("solver", ALL_SOLVERS, indirect=True)
def test_all_items_fit(solver):
    """Test scenario where all items fit in the knapsack."""
    items = np.array(
        [
            Item(value=10, weight=5),
            Item(value=20, weight=5),
            Item(value=30, weight=5),
        ]
    )
    capacity = 15
    solutions = solver(items, capacity)

    if not isinstance(solutions, list):
        solutions = [solutions]

    assert len(solutions) == 1
    assert solutions[0].value == 60
    assert solutions[0].weight == 15


@pytest.mark.parametrize("solver", ALL_SOLVERS, indirect=True)
def test_all_items_do_not_fit(solver):
    """Test scenario where no items fit in the knapsack."""
    items = np.array(
        [
            Item(value=10, weight=15),
            Item(value=20, weight=15),
            Item(value=30, weight=15),
        ]
    )
    capacity = 10
    solutions = solver(items, capacity)

    if not isinstance(solutions, list):
        solutions = [solutions]

    assert len(solutions) == 1
    assert solutions[0].value == 0
    assert solutions[0].weight == 0
    assert np.array_equal(solutions[0].state, np.zeros(len(items)))


@pytest.mark.parametrize("solver", OPTIMAL_SOLVERS, indirect=True)
@pytest.mark.parametrize("case", TEST_CASES)
def test_correct_optimal_found(solver, case):
    """Test that the correct optimal solution is found."""
    items = np.array(
        [
            Item(value=case["values"][i], weight=case["weights"][i])
            for i in range(len(case["values"]))
        ]
    )
    solution = solver(np.array(items), case["capacity"])
    if isinstance(solution, list):
        solution = solution[0]

    assert np.isclose(solution.value, case["optimal_value"])


@pytest.mark.parametrize("num_items", [5, 10, 15, 20])
@pytest.mark.parametrize("seed", [1, 2, 3, 4])
def test_branch_and_bound_decision_variant(num_items, seed):
    """Test that the branch-and-bound (decision) algorithm is correct."""
    sampler = Sampler(num_items=num_items, normalised_capacity=0.5)
    instance = sampler.sample(seed=seed)
    instance.solve()
    optimal = instance.optimal_nodes[0].value

    assert solvers._branch_and_bound_decision_variant(
        instance.items, instance.capacity, optimal
    )
    assert not solvers._branch_and_bound_decision_variant(
        instance.items, instance.capacity, optimal + 0.001
    )



"""
Provides an interface for defining arrangements of items.

Example:
    Define an arangement of items for the knapsack problem::

        from pykp import Arrangement, Item

        items = [
            Item(value=10, weight=5),
            Item(value=20, weight=10),
            Item(value=30, weight=15),
        ]
        state = [0, 1, 1]
        arrangement = Arrangement(items=items, state=state)
        print(arrangement)
"""

import numpy as np

from .item import Item


class Arrangement:
    """Represents an arrangement of items for the knapsack problem.

    Used to store the state of items in the knapsack, as well as the total
    value and weight of items in the arrangement. The state is represented
    as a binary array, where 1 indicates the item is in the arrangement, and
    0 indicates it is not.


    Parameters
    ----------
    items: list[Item]
        An array of items for the knapsack problem.
    state: list[int]
        Binary array indicating the inclusion or exclusion of items in the
        arrangement.
    value: int
        The total value of items in the arrangement.
    weight int:
        The total weight of items in the arrangement.
    """

    def __init__(
        self,
        items: list[Item],
        state: list[int],
    ):
        if not np.all(np.isin(state, [0, 1])):
            raise ValueError("Elements of `state` must be 0 or 1.")

        if isinstance(items, list):
            items = np.array(items)

        if isinstance(state, list):
            state = np.array(state)

        self._items = items
        self._state = state
        self._value = self.__calculate_value()
        self._weight = self.__calculate_weight()

    @property
    def items(self) -> list[Item]:
        """The items in the arrangement."""
        return list(self._items)

    @property
    def state(self) -> list[int]:
        """
        The state of the arrangement.

        1 indicates the item is in the arrangement, 0 indicates it is not.
        """
        return list(self._state)

    @property
    def value(self) -> float:
        """The total value of items in the arrangement."""
        return self._value

    @property
    def weight(self) -> float:
        """The total weight of items in the arrangement."""
        return self._weight

    def __calculate_value(self):
        """Calculate total value of items currently in the knapsack.

        Returns
        -------
        float
            The total value of items in the knapsack.
        """
        return sum(
            [
                self._items[i].value
                for i, inside in enumerate(self._state)
                if bool(inside)
            ]
        )

    def __calculate_weight(self):
        """Calculate the total weight of items currently in the knapsack.

        Returns
        -------
        float
            The total weight of items in the knapsack.
        """
        return sum(
            [
                self._items[i].weight
                for i, inside in enumerate(self._state)
                if bool(inside)
            ]
        )

    def __hash__(self):
        """Hash the state of the arrangement."""
        return hash(tuple(self._state))

    def __eq__(self, other):
        """Check if two arrangements are equal."""
        return np.array_equal(self._state, other.state)

    def __str__(self):
        """Return a string representation of the arrangement."""
        state = int("".join(self._state.astype(int).astype(str)), 2)
        return f"(v: {self._value}, w: {self._weight}, s: {state})"

    def __repr__(self):
        """Return a string representation of the arrangement."""
        state = int("".join(self._state.astype(int).astype(str)), 2)
        return f"(v: {self._value}, w: {self._weight}, s: {state})"



"""Tests for pykp.knapsack module."""

import pathlib

import numpy as np
import pytest

from pykp import Item, Knapsack


@pytest.fixture
def knapsack() -> Knapsack:
    """Return a knapsack instance with a set of items and capacity."""
    items = [
        Item(value=7, weight=3),
        Item(value=10, weight=5),
        Item(value=15, weight=10),
        Item(value=10, weight=10),
    ]
    capacity = 23
    return Knapsack(items=items, capacity=capacity)


def test_initialisation(knapsack: Knapsack):
    """Test the initialisation of a knapsack instance."""
    assert knapsack.capacity == 23
    assert np.array_equal(knapsack.state, np.zeros(len(knapsack.items)))
    assert knapsack.value == 0
    assert knapsack.weight == 0
    assert knapsack.is_feasible


def test_invalid_initialisation():
    """Test the initialisation of a knapsack instance with invalid inputs."""
    with pytest.raises(ValueError):
        Knapsack(items=np.array([]), capacity=10)

    with pytest.raises(ValueError):
        Knapsack(items="not an array", capacity=10)

    with pytest.raises(ValueError):
        Knapsack(items=np.array([1, 2, 3]), capacity=-5)


def test_add_item(knapsack: Knapsack):
    """Test adding an item to the knapsack."""
    knapsack.empty()
    knapsack.add(knapsack.items[0])
    assert knapsack.value == 7
    assert knapsack.weight == 3
    assert knapsack.is_feasible
    assert np.array_equal(knapsack.state, [1, 0, 0, 0])


def test_remove_item(knapsack: Knapsack):
    """Test removing an item from the knapsack."""
    knapsack.empty()
    knapsack.add(knapsack.items[0])
    knapsack.remove(knapsack.items[0])
    assert knapsack.value == 0
    assert knapsack.weight == 0
    assert np.array_equal(knapsack.state, [0, 0, 0, 0])


def test_set_state(knapsack: Knapsack):
    """Test setting the state of the knapsack."""
    knapsack.state = [1, 0, 1, 0]
    print(knapsack.items)
    assert knapsack.value == 22
    assert knapsack.weight == 13
    assert knapsack.is_feasible
    assert np.array_equal(knapsack.state, [1, 0, 1, 0])


def test_empty(knapsack: Knapsack):
    """Test emptying the knapsack."""
    knapsack.state = [1, 0, 1, 0]
    knapsack.empty()
    assert knapsack.value == 0
    assert knapsack.weight == 0
    assert np.array_equal(knapsack.state, [0, 0, 0, 0])


def test_load_from_json(tmp_path: pathlib.Path, knapsack: Knapsack):
    """Test loading a knapsack instance from a JSON file."""
    path = tmp_path / "test_knapsack.json"
    knapsack.write_to_json(str(path))
    new_knapsack = Knapsack(
        items=np.array([]), capacity=0, load_from_json=True, path_to_spec=path
    )
    assert new_knapsack.capacity == knapsack.capacity
    assert len(new_knapsack.items) == len(knapsack.items)



"""Interface for defining instances of the 0-1 Knapsack Problem.

Knapsack instances are defined by a set of items, each with weights and values,
and a capacity constraint. The :class:`Knapsack` class provides methods to
solve instances using different algorithms, add or remove items, and visualise
the solution.

Examples
--------
To define a Knapsack instance, initialise the `Knapsack` class with `Items`
and a capacity constraint

>>> from pykp import Knapsack
>>> from pykp import Item
>>> items = [
...     Item(value=10, weight=5),
...     Item(value=15, weight=10),
...     Item(value=7, weight=3),
... ]
>>> capacity = 15
>>> knapsack = Knapsack(items=items, capacity=capacity)
>>> knapsack.solve()
>>> print(knapsack.optimal_nodes)
[(v: 25, w: 15, s: 3)]
"""

import json
from typing import Literal, Union
from warnings import warn

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd

from .arrangement import Arrangement
from .item import Item
from .metrics import sahni_k
from .solvers import branch_and_bound, brute_force, mzn_gecode

SOLVERS = ["branch_and_bound", "mzn_gecode", "brute_force"]


class Knapsack:
    """Represents a knapsack problem solver.

    An instance is defined by a set of `items` and a `capacity` constraint.
    The `Knapsack` class provides methods to dynamically add or remove items,
    solve the knapsack problem using different algorithms, and visualise the
    solution.

    Parameters
    ----------
    items : list of Item
        A list of `Item` objects, each representing a candidate for the
        knapsack with associated value and weight.
    capacity : float
        The maximum total weight allowed in the knapsack.
    load_from_json : bool, optional
        Whether to load the items and capacity from a JSON file. If set to
        True, `path_to_spec` must be provided. Default is False.
    path_to_spec : str, optional
        Path to the JSON specification file containing an array of items
        (`value`, `weight`) and the knapsack capacity. Only used if
        `load_from_json` is True. Default is None.

    Examples
    --------
    Create a knapsack instance and solve using default settings:

    >>> from pykp import Knapsack
    >>> from pykp import Item
    >>> items = [
    ...     Item(value=10, weight=5),
    ...     Item(value=15, weight=10),
    ...     Item(value=7, weight=3),
    ... ]
    >>> capacity = 15
    >>> knapsack = Knapsack(items=items, capacity=capacity)
    >>> knapsack.solve()
    >>> print(knapsack.optimal_nodes)
    [(v: 25, w: 15, s: 3)]

    Solve the knapsack using the branch-and-bound method
    and explore the optimal nodes:

    >>> from knapsack_solver import Item, Knapsack
    >>> items = [
    ...     Item(value=10, weight=10),
    ...     Item(value=5, weight=5),
    ...     Item(value=5, weight=5),
    ... ]
    >>> knapsack = Knapsack(items=items, capacity=10)
    >>> knapsack.solve(method="branch_and_bound")
    >>> print(len(knapsack.optimal_nodes)
    2
    >>> for arrangement in knapsack.optimal_nodes:
    ...     print(
    ...         "Value:",
    ...         arrangement.value,
    ...         "Weight:",
    ...         arrangement.weight,
    ...         "State:",
    ...         arrangement.state,
    ...     )
    Value: 10 Weight: 10 State: [1.0 0.0 0.0]
    Value: 10 Weight: 10 State: [0.0 1.0 0.0]

    See Also
    --------
    pykp.item : Represents an item with value and weight attributes.
    pykp.solvers : Module containing solvers for the knapsack problem.
    """

    def __init__(
        self,
        items: list[Item],
        capacity: float,
        load_from_json: bool = False,
        path_to_spec: str = None,
    ):
        if load_from_json:
            self.load_from_json(path_to_spec)
            return

        if len(items) == 0:
            raise ValueError("`items` must have length greater than 0.")
        if not np.all([isinstance(item, Item) for item in items]):
            raise ValueError("All elements in `items` must be of type `Item`.")
        if capacity < 0:
            raise ValueError("`capacity` must be non-negative.")
        if not isinstance(items, np.ndarray):
            items = np.array(items)

        self._items = np.array(
            sorted(
                items, key=lambda item: item.value / item.weight, reverse=True
            )
        )
        self._capacity = capacity
        self._state = np.zeros_like(items)
        self._value = 0
        self._weight = 0
        self._is_feasible = True
        self._is_at_capacity = False

        self._nodes = np.array([])
        self._feasible_nodes = np.array([])
        self._terminal_nodes = np.array([])
        self._optimal_nodes = np.array([])

    @property
    def items(self) -> list[Item]:
        """The items in the knapsack."""
        return list(self._items)

    @property
    def capacity(self) -> float:
        """The capacity constraint of the knapsack."""
        return self._capacity

    @property
    def state(self) -> list[int]:
        """The binary array indicating the inclusion/exclusion of items."""
        return list(self._state)

    @state.setter
    def state(self, state: Union[list, np.ndarray]):
        """Set the knapsack state using the provided binary array.

        Parameters
        ----------
            state (list or np.ndarry): Binary array indicating the
                inclusion/exclusion of items in the knapsack.
        """
        if isinstance(state, list):
            state = np.array(state)
        self._state = state
        self.__update_state()

    @property
    def value(self) -> float:
        """The total value of items currently in the knapsack."""
        return self._value

    @property
    def weight(self) -> float:
        """The total weight of items currently in the knapsack."""
        return self._weight

    @property
    def is_feasible(self) -> bool:
        """Whether the knapsack is within its weight capacity."""
        return self._is_feasible

    @property
    def is_at_capacity(self) -> bool:
        """Whether the knapsack is at full capacity.

        A knapsack is at full capacity when placing in any item
        from the current set of excluded items would cause the sum of
        weights to exceed the capacity constraint.
        """
        return self._is_at_capacity

    @property
    def optimal_nodes(self) -> list[Arrangement]:
        """An array of optimal nodes in the knapsack.

        Optimal nodes are arrangements of items that maximise the total value
        of items in the knapsack, subject to the weight constraint.
        Optimal nodes are a subset of ``terminal_nodes``.
        """
        return list(self._optimal_nodes)

    @property
    def terminal_nodes(self) -> list[Arrangement]:
        """An array of terminal nodes in the knapsack.

        Terminal nodes are arrangements of items that are under the weight
        constraint, and at full capacity (no more items can be added without
        exceeding the capacity constraint). Terminal nodes are a subset of
        ``feasible_nodes``.
        """
        return list(self._terminal_nodes)

    @property
    def feasible_nodes(self) -> list[Arrangement]:
        """An array of feasible nodes in the knapsack.

        Feasible nodes are arrangements of items that are under the weight
        constraint.
        """
        return list(self._feasible_nodes)

    @property
    def nodes(self) -> list[Arrangement]:
        """An array of all nodes in the knapsack."""
        return list(self._nodes)

    def solve(
        self,
        method: Literal[
            "branch_and_bound", "mzn_gecode", "brute_force"
        ] = "branch_and_bound",
    ):
        """Solves the knapsack problem to find optimal arrangements.

        Parameters
        ----------
        method : {"branch_and_bound", "mzn_gecode", "brute_force"}, optional
            The algorithm to use for solving. Default is "branch_and_bound".

        Returns
        -------
        np.ndarray
            An array of `Arrangement` objects representing the optimal
            solutions.

        Raises
        ------
        ValueError
            If the specified `method` is invalid.

        Examples
        --------
        >>> from pykp import Knapsack
        >>> from pykp import Item
        >>> items = [
        ...     Item(value=10, weight=5),
        ...     Item(value=15, weight=10),
        ...     Item(value=7, weight=3),
        ... ]
        >>> capacity = 15
        >>> knapsack = Knapsack(items=items, capacity=capacity)
        >>> knapsack.solve(method="branch_and_bound")
        >>> print(knapsack.optimal_nodes)
        [(v: 25, w: 15, s: 3)]
        """
        if method not in SOLVERS:
            raise ValueError(f"`method` must be one of: {SOLVERS}.")

        if method == "branch_and_bound":
            self._optimal_nodes = branch_and_bound(
                items=self._items, capacity=self._capacity
            )

        if method == "mzn_gecode":
            result = mzn_gecode(items=self._items, capacity=self._capacity)
            self._optimal_nodes = list(result)

        if method == "brute_force":
            if len(self._items) > 15:
                warn(
                    message="Brute force is infeasible for large instances.",
                    category=RuntimeWarning,
                    stacklevel=2,
                )
            (
                self._optimal_nodes,
                self._terminal_nodes,
                self._feasible_nodes,
                self._nodes,
            ) = brute_force(items=self._items, capacity=self._capacity)

        return self.optimal_nodes

    def add(self, item: Item) -> list:
        """
        Includes a specific item in the knapsack.

        Parameters
        ----------
        item : Item
            The item to add. Must already exist among `self.items`.

        Returns
        -------
        list
            The updated binary state array after adding the item.

        Raises
        ------
        ValueError
            If `item` is not of type `Item` or if it is not present in
            `self.items`.

        Examples
        --------
        >>> items = [Item(10, 5), Item(15, 5)]
        >>> knapsack = Knapsack(
        >>>     items=items,
        ...     capacity=6
        ... )
        >>> knapsack.add(items[1])
        array([0, 1])
        >>> knapsack.value
        15
        >>> knapsack.weight
        5
        """
        if not isinstance(item, Item):
            raise ValueError("`item` must be of type `Item`.")
        if item not in self._items:
            raise ValueError(
                """`item` must be an existing `item` inside the `Knapsack`
                instance."""
            )
        self._state[np.where(self._items == item)[0][0]] = 1
        self.__update_state()
        return self.state

    def remove(self, item: Item) -> list:
        """
        Remove a specific item from the knapsack.

        Parameters
        ----------
        item : Item
            The item to remove. Must already exist among `self.items`.

        Returns
        -------
        list
            The updated binary state array after removing the item.

        Raises
        ------
        ValueError
            If `item` is not of type `Item` or if it is not present in
            `self.items`.

        Examples
        --------
        >>> items = [Item(10, 5), Item(15, 5)]
        >>> knapsack = Knapsack(
        >>>     items=items,
        ...     capacity=6
        ... )
        >>> knapsack.add(items[1])
        array([0, 1])
        >>> knapsack.remove(items[1])
        array([0, 0])
        >>> knapsack.value
        0
        >>> knapsack.weight
        0
        """
        if not isinstance(item, Item):
            raise ValueError("`item` must be of type `Item`.")
        if item not in self._items:
            raise ValueError(
                """`item` must be an existing `item` inside the `Knapsack`
                instance."""
            )

        self._state[np.where(self._items == item)] = 0
        self.__update_state()
        return self.state

    def empty(self):
        """Remove all items from the knapsack.

        Returns
        -------
        list
            The updated binary state array, all set to 0.

        Examples
        --------
        >>> knapsack = Knapsack(items=[Item(10, 2), Item(15, 4)], capacity=6)
        >>> knapsack.add(knapsack.items[0])
        array([1, 0])
        >>> knapsack.empty()
        array([0, 0])
        """
        self._state = np.zeros_like(self._items)
        self.__update_state()
        return self.state

    def __update_state(self):
        """Update the internal state of the knapsack."""
        self._value = self.__calculate_value()
        self._weight = self.__calculate_weight()
        self._is_feasible = self._capacity >= self._weight
        out_items = [
            self._items[i]
            for i, element in enumerate(self._state)
            if element == 0
        ]
        if sum(self._state) == len(self._state):
            self._is_at_capacity = True
        else:
            self._is_at_capacity = (
                min([self._weight + item.weight for item in out_items])
                > self._capacity
            )

    def __calculate_value(self) -> float:
        """Calculate the total value of items included in the current state.

        Returns
        -------
        float
            Sum of the values of included items.
        """
        mask = np.ma.make_mask(self._state, shrink=False)
        return sum([item.value for item in self._items[mask]])

    def __calculate_weight(self):
        """Calculate the total weight of items included in the current state.

        Returns
        -------
        float
            Sum of the weights of included items.
        """
        mask = np.ma.make_mask(self._state, shrink=False)
        return sum([item.weight for item in self._items[mask]])

    def plot_terminal_nodes_histogram(
        self, ax: plt.Axes = None
    ) -> tuple[plt.Figure, plt.Axes]:
        """Plot a histogram of terminal node values.

        All nodes are enumerated to find terminal nodes, and then plotted
        as a histogram. Terminal nodes are arrangements of items that are
        under the weight constraint, and at full capacity––that is, no more
        items can be added without exceeding the capacity constraint. Terminal
        nodes are a subset of ``feasible_nodes``.

        If the terminal nodes have not been enumerated, this method will first
        call `solve` with the `brute_force` method to find the terminal nodes.
        Note that this is infeasible for large n.

        Parameters
        ----------
        ax : matplotlib.pyplot.Axes, optional
            The axes object to plot the histogram. Defaults to None.

        Returns
        -------
        tuple of (matplotlib.pyplot.Figure, matplotlib.pyplot.Axes)
            The figure and axes objects of the resulting histogram.

        Examples
        --------
        >>> import random
        >>> from pykp import Item, Knapsack
        >>> import matplotlib.pyplot as plt
        >>>
        >>> random.seed(42)
        >>> weights = [random.randint(1, 100) for _ in range(12)]
        >>> values = [random.randint(1, 100) for _ in range(12)]
        >>> items = [Item(v, w) for v, w in zip(values, weights)]
        >>> capacity = sum(weights) / 2
        >>> knapsack = Knapsack(items=items, capacity=capacity)
        >>> knapsack.solve(method="brute_force")
        [(v: 372, w: 300, s: 4000)]
        >>> fig, ax = knapsack.plot_terminal_nodes_histogram()
        >>> plt.show()

        .. image:: /_static/plots/terminal_nodes_hist.png
            :alt: Histogram of terminal node values
        """
        if not len(self.nodes) == 2 ** len(self.items):
            self.solve(method="brute_force")

        if not ax:
            fig, ax = plt.subplots(nrows=1, ncols=1, constrained_layout=True)

        ax.hist(
            [arrangement.value for arrangement in self._terminal_nodes],
            bins=100,
            alpha=1,
        )
        ax.set_ylabel("Number of terminal nodes")
        ax.set_xlabel("Value")

        return fig, ax

    def __get_node_color(self, arrangement):
        """Get the colour of a node in the network plot."""
        # Optimal node
        if arrangement.value == self._optimal_nodes[0].value:
            return "#57ff29"

        # Feasible nodes
        if arrangement.weight < self._capacity:
            return "#003CAB"

        # Infeasible nodes
        return "#FF2C00"

    def plot_network(
        self,
        ax: plt.Axes = None,
    ) -> tuple[plt.Figure, plt.Axes]:
        """
        Visualises a graph representation of the knapsack problem.

        Each node in the graph represents a unqiue arrangement of items inside
        the knapsack. Edges connect nodes whose states differ by a single
        item,  and represent the elementary operations of adding or removing a
        single item from the knapsack.

        Parameters
        ----------
        ax : matplotlib.pyplot.Axes, optional
            An existing axes object to plot on. Defaults to None.

        Returns
        -------
        tuple of (matplotlib.pyplot.Figure, matplotlib.pyplot.Axes)
            The figure and axes of the created or updated plot.

        Examples
        --------
        >>> import random
        >>> from pykp import Item, Knapsack
        >>> import matplotlib.pyplot as plt
        >>>
        >>> random.seed(42)
        >>> weights = [random.randint(1, 100) for _ in range(5)]
        >>> values = [random.randint(1, 100) for _ in range(5)]
        >>> items = [Item(v, w) for v, w in zip(values, weights)]
        >>> capacity = sum(weights) / 2
        >>> knapsack = Knapsack(items=items, capacity=capacity)
        >>> knapsack.solve(method="brute_force")
        [(v: 142, w: 114, s: 28)]
        >>> fig, ax = knapsack.plot_network()
        >>> plt.show()

        .. image:: /_static/plots/network.png
            :alt: Knapsack network representation
        """
        if not len(self.nodes) == 2 ** len(self.items):
            self.solve_all_nodes()

        kp_network = kp_network = nx.DiGraph()

        for arrangement in self._nodes:
            neighbours = [
                alt_arrangement
                for alt_arrangement in self._nodes
                if np.sum(
                    np.abs(
                        np.subtract(alt_arrangement.state, arrangement.state)
                    )
                )
                == 1
            ]

            kp_network.add_node(
                arrangement,
            )
            kp_network.add_edges_from(
                [
                    (arrangement, alt_arrangement)
                    for alt_arrangement in neighbours
                ]
            )

        if ax is None:
            fig, ax = plt.subplots(
                figsize=(4 * len(self._items) / 10, 4 * len(self._items) / 10),
                dpi=200,
                nrows=1,
                ncols=1,
                constrained_layout=True,
            )

        node_colors = [
            self.__get_node_color(arrangement)
            for arrangement in kp_network.nodes
        ]
        nx.draw_spring(
            kp_network,
            ax=ax,
            node_color=node_colors,
            node_size=2,
            width=0.05,
            arrowsize=0.01,
            with_labels=False,
        )

        return fig, ax

    def write_to_json(self, path: str):
        """Write the knapsack configuration to a JSON file.

        The output file can be used to load the knapsack instance at a later
        time using the `load_from_json` method.

        Parameters
        ----------
        path : str
            The file path for the output JSON file.

        Examples
        --------
        >>> knapsack = Knapsack(items=[Item(10, 2), Item(20, 4)], capacity=6)
        >>> knapsack.solve()
        >>> knapsack.write_to_json("output.json")
        """
        instance_spec = {
            "capacity": self._capacity,
            "items": [
                {
                    "id": i,
                    "value": item.value,
                    "weight": item.weight,
                }
                for i, item in enumerate(self._items)
            ],
        }

        with open(path, "w") as f:
            json.dump(instance_spec, f, indent=4, default=int)

    def load_from_json(self, path: str):
        """Load a knapsack configuration from a JSON file.

        Parameters
        ----------
        path : str
            The file path of the JSON specification.

        Examples
        --------
        >>> knapsack = Knapsack(
        ...     items=[],
        ...     capacity=0,
        ...     load_from_json=True,
        ...     path_to_spec="knapsack_config.json",
        ... )
        """
        with open(path) as f:
            spec = json.load(f)
            self.__init__(
                items=np.array(
                    [
                        Item(item["value"], item["weight"])
                        for item in spec["items"]
                    ]
                ),
                capacity=int(spec["capacity"]),
            )

    def summary(self):
        """Return a DataFrame summarising the knapsack instance.

        Returns
        -------
        pandas.DataFrame
            A summary table describing the knapsack instance.

        Examples
        --------
        >>> knapsack = Knapsack(items=[Item(10, 2), Item(20, 4)], capacity=6)
        >>> knapsack.solve(method="brute_force")
        >>> df = knapsack.summary()
        >>> print(df)
        """
        n_terminal = 2 ** len(self._items)
        n_optimal = len(self._optimal_nodes)

        header = [
            f"C = {self._capacity}",
            f"nC = {
                round(
                    self._capacity
                    / np.sum([item.weight for item in self._items]),
                    2,
                )
            }",
            f"nTerminal = {n_terminal}",
            f"nOptimal = {n_optimal}",
        ]

        if len(self._terminal_nodes) > len(self._optimal_nodes):
            best_inferior_solution = self._terminal_nodes[
                len(self._optimal_nodes)
            ]
            delta = self._optimal_nodes[0].value - best_inferior_solution.value
            delta_pct = delta / self._optimal_nodes[0].value
            header.append(f"Δ = {delta}")
            header.append(f"Δ% = {delta_pct:.3}")
        else:
            best_inferior_solution = None

        header = ", ".join(header)

        columns = pd.MultiIndex.from_arrays(
            [
                [header] * len(self._items),
                [i + 1 for i, item in enumerate(self._items)],
            ]
        )

        rows = [
            [item.value for item in self._items],
            [item.weight for item in self._items],
            [round(item.value / item.weight, 3) for item in self._items],
        ]
        rows.extend(
            [
                np.where(arrangement.state == 1, "IN", "OUT")
                for arrangement in self._optimal_nodes
            ]
        )

        index = ["v", "w", "density"]
        index.extend(
            [
                ", ".join(
                    [
                        f"solution (v = {arrangement.value}",
                        f"w = {arrangement.weight}",
                        f"k = {sahni_k(arrangement, self._capacity)})",
                    ]
                )
                for arrangement in self._optimal_nodes
            ]
        )
        if best_inferior_solution is not None:
            index.append(
                ", ".join(
                    [
                        f"best inferior (v = {best_inferior_solution.value}",
                        f"w = {best_inferior_solution.weight}",
                        f"k = {
                            sahni_k(best_inferior_solution, self._capacity)
                        })",
                    ]
                )
            )
            rows.append(
                np.where(best_inferior_solution.state == 1, "IN", "OUT")
            )

        return pd.DataFrame(rows, columns=columns, index=index, dtype="object")

    def __str__(self):
        """Return a string representation of the knapsack instance."""
        return repr(self.summary())



"""Tests for pykp.metrics module."""

import os
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest

import pykp.metrics as metrics

SOLVERS = ["branch_and_bound"]


@pytest.mark.parametrize("outcome", ["solvability", "time", "both"])
@pytest.mark.parametrize("resolution", [(5, 5), (2, 3), (10, 10)])
def test_phase_transition_returns_correct_shape(resolution, outcome):
    """Test dimensions of grid and phase transition match resolution."""
    grid, phase_transition = metrics.phase_transition(
        num_items=5,
        samples=2,
        outcome=outcome,
        solver="branch_and_bound",
        resolution=resolution,
        path=None,
    )
    assert grid[0].shape == resolution
    assert grid[1].shape == resolution

    if outcome == "both":
        assert phase_transition[0].shape == resolution
        assert phase_transition[1].shape == resolution
    else:
        assert phase_transition.shape == resolution


@pytest.mark.parametrize("solver", SOLVERS)
def test_phase_transition_with_solvers(solver):
    """Test that all solvers do not raise an error."""
    try:
        metrics.phase_transition(
            num_items=5,
            samples=2,
            solver=solver,
            resolution=(2, 2),
            path=None,
        )
    except Exception as e:
        pytest.fail(f"Unexpected error occurred for solver: {e}")


def test_phase_transition_with_invalid_solver_raises():
    """Test that passing an invalid solver raises a ValueError."""
    with pytest.raises(ValueError):
        metrics.phase_transition(
            num_items=5, samples=2, solver="invalid_solver", resolution=(2, 2)
        )


@pytest.mark.parametrize("outcome", ["time", "solvability", "both"])
def test_phase_transition_saves_csv(tmp_path, outcome):
    """Test saving .csv file."""
    output_path = tmp_path / "phase_transition_output.csv"
    resolution = (2, 2)

    grid, solvability_matrix = metrics.phase_transition(
        num_items=5,
        samples=2,
        outcome=outcome,
        solver="branch_and_bound",
        resolution=resolution,
        path=str(output_path),
    )

    assert output_path.exists(), "CSV file was not created."

    df = pd.read_csv(output_path)
    expected_columns = [
        "nc_lower",
        "nc_upper",
        "np_lower",
        "np_upper",
    ]
    if outcome == "time":
        expected_columns.append("time")
    elif outcome == "solvability":
        expected_columns.append("solvability")
    elif outcome == "both":
        expected_columns.extend(["time", "solvability"])
    assert all(col in df.columns for col in expected_columns), (
        "Missing expected columns in CSV."
    )

    # Check row count = resolution[0] * resolution[1]
    assert len(df) == (resolution[0] * resolution[1]), (
        "CSV file has unexpected number of rows."
    )


def test_phase_transition_values_in_range():
    """Test that the values in the solvability matrix lie between 0 and 1."""
    resolution = (2, 2)
    grid, solvability_matrix = metrics.phase_transition(
        num_items=5,
        samples=2,
        outcome="solvability",
        solver="branch_and_bound",
        resolution=resolution,
        path=None,
    )
    assert np.all(solvability_matrix >= 0.0) and np.all(
        solvability_matrix <= 1.0
    ), "Solvability values are not all within [0, 1]."


@pytest.mark.parametrize("seed", [1, 2])
def test_phase_transition_reproducibility(seed):
    """Test that the phase transition is reproducible."""
    resolution = (2, 2)
    num_items = 5
    samples = 5
    grid, solvability_matrix = metrics.phase_transition(
        num_items=num_items,
        samples=samples,
        outcome="solvability",
        solver="branch_and_bound",
        resolution=resolution,
        path=None,
        seed=seed,
    )

    grid2, solvability_matrix2 = metrics.phase_transition(
        num_items=num_items,
        samples=samples,
        outcome="solvability",
        solver="branch_and_bound",
        resolution=resolution,
        path=None,
        seed=seed,
    )

    assert np.all(grid[0] == grid2[0]), "Inconsistent grid."
    assert np.all(grid[0] == grid2[0]), "Inconsistent grid."
    assert np.all(solvability_matrix == solvability_matrix2), (
        "Inconsistent solvability matrix."
    )



"""Interface for defining items inside a knapsack problem.

The :class:`Item` class is used to define items for the knapsack problem. Each
item has a value and a weight associated with it.

Example:
    Use the Item class to define items for the knapsack problem::

        from pykp import Item

        items = [
            Item(value=10, weight=5),
            Item(value=20, weight=10),
            Item(value=30, weight=15),
        ]

"""

from dataclasses import dataclass, field


@dataclass(frozen=True, eq=False)
class Item:
    """
    Represents an item for the knapsack problem.

    Parameters
    ----------
        value: int
            The value of the item.
        weight: int
            The weight of the item.
    """

    value: int = field(compare=False)
    weight: int = field(compare=False)
